# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_grids_gridspec.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.GridSpec instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.GridSpec

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Grid Spec"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.grids.GridSpec.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = '56259768-e2b1-11df-aab5-00163e9152a5'

# Test document version.
DOC_VERSION = '1'

# Test document creation date.
DOC_DATE = arrow.get('2012-04-23 14:59:07.384733').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_extension_info(ext):
    """Asserts a document's extension information.

    :param object ext: Document extension information.

    """
    tu.assert_str(ext.display_name, "UM N96L38 ATM Grid System")
    tu.assert_str(ext.description, "Met Office Unified Model", True)
    tu.assert_str(ext.full_display_name, "CMIP5 Grid Spec : MOHC - UM N96L38 ATM Grid System")
    tu.assert_int(ext.summary_fields, 2)
    tu.assert_str(ext.summary_fields[0], "UM N96L38 ATM Grid System")
    tu.assert_str(ext.summary_fields[1], "Met Office Unified Model", True)


def _assert_doc_core(doc, is_update):
    """Assert core information."""
    tu.assert_iter(doc.esm_model_grids, 1, cim.v1.GridMosaic)
    g = doc.esm_model_grids[0]
    tu.assert_iter(g.citations, 2, cim.v1.Citation)
    tu.assert_str(g.description, "Specification of the atmosphere", True)
    tu.assert_bool(g.is_leaf, True)
    tu.assert_str(g.long_name, "Met Office Unified Model 192", True)
    tu.assert_str(g.short_name, "UM N96L38 ATM Grid System")
    tu.assert_str(g.type, "regular_lat_lon")


def _assert_doc_tile(doc, is_update):
    """Assert tile information."""
    g = doc.esm_model_grids[0]
    tu.assert_int(g.tile_count, 1)
    tu.assert_iter(g.tiles, 1, cim.v1.GridTile)
    tu.assert_object(g.tiles[0], cim.v1.GridTile)
    t = g.tiles[0]
    tu.assert_str(t.description, "Horizontal properties: The N96", True)
    tu.assert_str(t.discretization_type, "logically_rectangular")
    tu.assert_str(t.mnemonic, "N96")


def _assert_doc_tile_extent(doc, is_update):
    """Assert tile extent information."""
    t = doc.esm_model_grids[0].tiles[0]
    tu.assert_object(t.extent, cim.v1.GridExtent)
    e = t.extent
    tu.assert_int(e.maximum_latitude, 90)
    tu.assert_int(e.minimum_latitude, -90)
    tu.assert_int(e.minimum_longitude, 0)
    tu.assert_int(e.maximum_longitude, 360, int)


def _assert_doc_tile_horizontal_resolution(doc, is_update):
    """Assert tile horizontal resolution."""
    t = doc.esm_model_grids[0].tiles[0]
    tu.assert_object(t.horizontal_resolution, cim.v1.GridTileResolutionType)
    hr = t.horizontal_resolution
    tu.assert_str(hr.description, "1.875 degrees in longitude", True)
    tu.assert_iter(hr.properties, 2, cim.v1.GridProperty)
    p = hr.properties[0]
    tu.assert_str(p.name, "NumberOfLatitudinalGridCells")
    tu.assert_str(p.value, "145")


def _assert_doc_tile_vertical_resolution(doc, is_update):
    """Assert tile vertical resolution."""
    t = doc.esm_model_grids[0].tiles[0]
    tu.assert_object(t.vertical_resolution, cim.v1.GridTileResolutionType)
    vr = t.vertical_resolution
    tu.assert_iter(vr.properties, 4, cim.v1.GridProperty)
    p = vr.properties[0]
    tu.assert_str(p.name, "TopModelLevel")
    tu.assert_str(p.value, "39254.8")


def _assert_doc_tile_zcoords(doc, is_update):
    """Assert tile zcoords information."""
    t = doc.esm_model_grids[0].tiles[0]
    tu.assert_object(t.zcoords, cim.v1.VerticalCoordinateList)
    z = t.zcoords
    tu.assert_str(z.form, "hybrid height")
    tu.assert_str(z.type, "hybrid")
    tu.assert_iter(z.properties, 1, cim.v1.GridProperty)


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_core,
        _assert_doc_tile,
        _assert_doc_tile_extent,
        _assert_doc_tile_horizontal_resolution,
        _assert_doc_tile_vertical_resolution,
        _assert_doc_tile_zcoords
        ):
        assertor(doc, is_update)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass


# -*- coding: utf-8 -*-

"""
.. module:: test_extensions.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Executes pyesdoc document extension tests.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
# Module imports.
import nose

import pyesdoc
import test_utils as tu
import test_types as tt



def _test_custom_extender_count():
    """Test number of extenders."""
    tu.assert_iter(pyesdoc.extensions.SUPPORTED, length=10)


def _test_is_custom_extendable():
    """Test extenders correctly mapped to document types."""
    for doc_type in pyesdoc.ontologies.get_types():
        is_extendable = doc_type.type_key.lower() in pyesdoc.extensions.SUPPORTED
        assert pyesdoc.is_extendable(doc_type.type_key) == is_extendable


def _test(mod):
    """Performs standard document extensions tests,"""
    doc = tu.get_doc(mod)    # Note - returns an extended document
    tu.assert_object(doc.ext)
    tu.assert_str(doc.type_key, doc.meta.type)
    tu.assert_str(doc.type_key, doc.ext.type)
    if hasattr(mod, "assert_doc_extensions"):
        mod.assert_doc_extensions(doc)


def test():
    """Performs extension tests over the set of test documents."""
    for func in (
        _test_custom_extender_count,
        _test_is_custom_extendable,
        ):
        tu.init(func, func.__doc__[5:])
        yield func

    for mod in tt.MODULES:
        tu.init(_test, 'extensions', mod)
        yield _test, mod



# -*- coding: utf-8 -*-

"""
.. module:: grids_grid_spec.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.grids.grid_spec document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_display_name,
        _set_description,
        _set_ids,
        _set_summary_fields,
        )


def _set_display_name(ctx):
    """Sets document display name.

    """
    if len(ctx.doc.esm_model_grids):
        ctx.ext.display_name = ctx.doc.esm_model_grids[0].short_name
    elif len(ctx.doc.esm_exchange_grids):
        ctx.ext.display_name = ctx.doc.esm_exchange_grids[0].short_name


def _set_description(ctx):
    """Sets document description.

    """
    if len(ctx.doc.esm_model_grids):
        ctx.ext.description = ctx.doc.esm_model_grids[0].long_name
    elif len(ctx.doc.esm_exchange_grids):
        ctx.ext.description = ctx.doc.esm_exchange_grids[0].long_name


def _set_ids(ctx):
    """Sets document id.

    """
    for grid in ctx.doc.esm_model_grids + ctx.doc.esm_exchange_grids:
        if not grid.id:
            grid.id = grid.short_name


def _set_summary_fields(ctx):
    """Sets document summary fields.

    """
    if len(ctx.doc.esm_model_grids):
    	ctx.ext.summary_fields += (
            ctx.doc.esm_model_grids[0].short_name,
            ctx.doc.esm_model_grids[0].long_name
        )
    elif len(ctx.doc.esm_exchange_grids):
    	ctx.ext.summary_fields += (
            ctx.doc.esm_exchange_grids[0].short_name,
            ctx.doc.esm_exchange_grids[0].long_name
        )



"""
.. module:: utils_loader.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Loads CMIP6 specialization from file system.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import imp
import json
import os
import glob



def get_short_tables_definitions(input_dir, typeof):
    """Returns specialization short tables.

    :param str input_dir: Directory within which modules reside.
    :param str typeof: Type of specialization being processed.

    :returns: Tuple of decoded short tables.
    :rtype: tuple

    """
    def _decode(fpath):
        name = fpath.split("/")[-1].split(".")[0].replace("_short_table", "")
        with open(fpath, 'r') as fstream:
            return name, json.loads(fstream.read())

    fpaths = os.path.join(input_dir, "short_tables")
    fpaths = os.path.join(fpaths, "{}_*.json".format(typeof))
    tables = [_decode(i) for i in glob.glob(fpaths)]

    return sorted(tables, key=lambda i: i[0])


def get_modules(input_dir, typeof):
    """Returns specialization modules.

    :param str input_dir: Directory within which modules reside.
    :param str typeof: Type of specialization being processed.

    """
    # Load specialization modules.
    modules = _get_modules(input_dir, typeof)
    if not modules:
        raise KeyError("Specializations not found")

    # Set specializations.
    root = _get_module(modules, typeof)
    try:
        root.GRID
    except AttributeError:
        grid = None
    else:
        grid = _get_module(modules, root.GRID)
    key_properties = _get_module(modules, root.KEY_PROPERTIES)
    processes = [_get_module(modules, p) for p in root.PROCESSES]

    return root, grid, key_properties, processes


def _get_modules(input_dir, specialization_type):
    """Returns a set of specialization modules.

    """
    modules = sorted([i for i in os.listdir(input_dir) if _is_target(i, specialization_type)])
    modules = [os.path.join(input_dir, m) for m in modules]
    modules = [(m.split("/")[-1].split(".")[0], m) for m in modules]

    return [imp.load_source(name, fpath) for name, fpath in modules]


def _get_module(modules, name):
    """Returns a specialization module.

    """
    for module in modules:
        if module.__name__ == name:
            return module


def _is_target(filename, specialization_type):
    """Returns flag indicating whether a module is a specialization target or not.

    """
    return not filename.startswith('_') and \
           filename.endswith('.py') and \
           filename.startswith(specialization_type)



# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_activity_numerical_experiment.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.NumericalExperiment instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.NumericalExperiment

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Experiment"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key

# Test representation file.
DOC_FILE = 'cim.1.activity.NumericalExperiment.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = '9fa513fc-d3a5-11df-837f-00163e9152a5'

# Test document version.
DOC_VERSION = '1'

# Test document creation date.
DOC_DATE = arrow.get('2012-04-23 14:59:07.048869').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def assert_extension_info(ext):
    """Asserts a document's extension information.

    :param object ext: Document extension information.

    """
    tu.assert_str(ext.display_name, "amip")
    tu.assert_str(ext.description, "AMIP (1979 - at least 2008)", True)
    tu.assert_str(ext.full_display_name, "CMIP5 Experiment : MOHC - amip")
    tu.assert_int(ext.summary_fields, 2)
    tu.assert_str(ext.summary_fields[0], "amip")
    tu.assert_str(ext.summary_fields[1], "AMIP")


def _assert_doc_core(doc, is_update):
    """Assert core information."""
    tu.assert_str(doc.description, "AMIP (1979 - at least 2008)", True)
    tu.assert_str(doc.experiment_id, "3.3")
    tu.assert_str(doc.long_name, "AMIP")
    tu.assert_iter(doc.rationales, 1, str)
    tu.assert_str(doc.rationales[0], "AMIP. Baseline simulation for", True)
    tu.assert_str(doc.short_name, "amip")


def _assert_doc_requirements(doc, is_update):
    """Assert experiment requirements."""
    tu.assert_iter(doc.requirements, 13, cim.v1.NumericalRequirement)

    # Boundary condition.
    req = doc.requirements[0]
    tu.assert_object(req, cim.v1.BoundaryCondition)
    tu.assert_str(req.requirement_type, "boundaryCondition")
    tu.assert_str(req.description, "Imposed changing concentrations", True)
    tu.assert_none(req.id)
    tu.assert_str(req.name, "3.3.bc.sls")

    # Boundary condition - options.
    tu.assert_iter(req.options, 2, cim.v1.NumericalRequirementOption)
    opt = req.options[0]
    tu.assert_str(opt.description, "Concentrations")
    tu.assert_str(opt.id, "bc.013")
    tu.assert_str(opt.name, "3.3.bc.sls_conc")
    tu.assert_str(opt.relationship, "XOR")

    # Initial condition.
    req = doc.requirements[11]
    tu.assert_object(req, cim.v1.InitialCondition)
    tu.assert_str(req.description, "Initial conditions are from", True)
    tu.assert_str(req.id, "ic.003")
    tu.assert_str(req.name, "3.3.ic")
    tu.assert_str(req.requirement_type, "initialCondition")

    # Spatio Temporal Constraint.
    req = doc.requirements[12]
    tu.assert_object(req, cim.v1.SpatioTemporalConstraint)
    tu.assert_object(req.date_range, cim.v1.ClosedDateRange)
    tu.assert_str(req.date_range.duration, "P30Y")
    tu.assert_date(req.date_range.end, "2008-12-30 00:00:00+00:00")
    tu.assert_date(req.date_range.start, "1979-01-01 00:00:00+00:00")
    tu.assert_str(req.description, "Begin in 1979 and", True)
    tu.assert_str(req.id, "stc.027")
    tu.assert_str(req.name, "3.3.stc.1979_30yr")
    tu.assert_str(req.requirement_type, "spatioTemporalConstraint")


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_core,
        _assert_doc_requirements,
        ):
        assertor(doc, is_update)


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass


"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface river routing'

# --------------------------------------------------------------------
# RIVER ROUTNG: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'TODO',
    'properties' : [
        ('tiling', 'str', '0.1',
             'Describe the river routing, if any.'),
        ('time_step', 'int', '1.1',
             'Time step of river routing scheme in seconds'),        
        ('grid_inherited_from_land_surface', 'bool', '1.1',
             'Is the grid inherited from land surface?'),
        ('grid_description', 'str', '0.1',
             'General description of grid, if not inherited from land surface'),
        ('number_of_reservoirs', 'int', '1.1',
             'Enter the number of reservoirs'),
        ('water_re_evaporation', 'ENUM:water_re_evaporation_types', '1.N',
             'TODO'),
        ('coupled_to_atmosphere', 'bool', '0.1',
             'Is river routing coupled to the atmosphere model component?'),
        ('coupled_to_land', 'str', '0.1',
             'Describe the coupling between land and rivers'),
        ('quantities_exchanged_with_atmosphere', 'ENUM:quantities_exchanged_with_atmosphere_types', '0.N',
             'If couple to atmosphere, which quantities are exchanged between river routing and the atmosphere model components?'),
        ('basin_flow_direction_map', 'ENUM:basin_flow_direction_map_types', '1.1',
             'What type of basin flow direction map is being used?'),
        ('flooding', 'str', '0.1',
             'Describe the representation of flooding, if any'),
        ('prognostic_variables', 'str', '1.1',
             'List the prognostic variables of the river routing'),
    ],
}

# --------------------------------------------------------------------
# RIVER ROUTING: process
# --------------------------------------------------------------------
DETAILS['oceanic_discharge'] = {
    'description': 'TODO',
    'properties' : [
        ('discharge_type', 'ENUM:discharge_types', '1.1',
             'Specify how rivers are discharged to the ocean'),
        ('quantities_transported', 'ENUM:quantities_transported_types', '1.N',
             'Quantities that are exchanged from river-routing to the ocean model component'),
    ]
}

# --------------------------------------------------------------------
# RIVER ROUTING: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['prognostic_variable_types'] = {
    'description': 'Specify the prognostic variables within the river routing scheme',
    'is_open': True,
    'members': [
        ('heat', None),
        ('water', None),
        ('tracers', None),
        ]
    }

ENUMERATIONS['water_re_evaporation_types'] = {
    'description': 'TODO',
    'is_open': True,
    'members': [
        ('flood plains', None),
        ('irrigation', None),
    ]
}

ENUMERATIONS['quantities_exchanged_with_atmosphere_types'] = {
    'description': 'Quantities that are exchanged between river routing and the atmosphere model components',
    'is_open': True,
    'members': [
        ('heat', None),
        ('water', None),
        ('tracers', None),
    ]
}

ENUMERATIONS['basin_flow_direction_map_types'] = {
    'description': 'What type of basin flow direction map is being used?',
    'is_open': True,
    'members': [
        ('present day', None),
        ('adapted for other periods', None),
    ]
}

ENUMERATIONS['discharge_types'] = {
    'description': 'Specify how rivers are discharged to the ocean',
    'is_open': True,
    'members': [
        ('direct (large rivers)', None),
        ('diffuse', None),
    ]
}

ENUMERATIONS['quantities_transported_types'] = {
    'description': 'Quantities that are exchanged from river-routing to the ocean model component',
    'is_open': True,
    'members': [
        ('heat', None),
        ('water', None),
        ('tracers', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.decoder_for_misc_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_xml_utils import set_attributes
from decoder_for_activity_package import *
from decoder_for_data_package import *
from decoder_for_grids_package import *
from decoder_for_shared_package import *
from decoder_for_software_package import *
import typeset



def decode_document_set(xml, nsmap):
    """Decodes an instance of the following type: document set.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.misc.DocumentSet

    """
    decodings = [
        ('data', True, decode_data_object, 'child::cim:dataObject'),
        ('ensembles', True, decode_ensemble, 'child::cim:ensemble'),
        ('experiment', False, decode_numerical_experiment, 'child::cim:numericalExperiment'),
        ('grids', True, decode_grid_spec, 'child::cim:gridSpec'),
        ('model', False, decode_model_component, 'child::cim:modelComponent'),
        ('platform', False, decode_platform, 'child::cim:platform'),
        ('simulation', False, decode_simulation_run, 'child::cim:simulationRun'),
    ]

    return set_attributes(typeset.misc.DocumentSet(), xml, nsmap, decodings)





# -*- coding: utf-8 -*-

"""
.. module:: drs_entities.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def drs_atomic_dataset():
    """An entity in a DRS file system.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('frequency', 'drs.drs_frequency_types', '1.1',
                "Frequency at which data is stored."),
            ('geographical_constraint', 'drs.drs_geographical_indicator', '0.1',
                "Identifies geographical subsets and spatial means."),
            ('mip_table', 'str', '1.1',
                "The MIP table, together with the variable defines the physical quantity."),
            ('realm', 'drs.drs_realms', '0.1',
                "Modelling realm."),
            ('temporal_constraint', 'drs.drs_temporal_identifier', '0.1',
                "Identifies temporal subsets or means."),
            ('variable_name', 'str', '1.1',
                "Identifies the physical quantity (when used in conjunction with the MIP table)."),
            ('version_number', 'int', '0.1',
                "Uniquely identifies a particular version of a publication level dataset.")
        ]
    }


def drs_ensemble_identifier():
    """Identifies a 'response ensemble' realisation using the semantic content ofa 'run_variant_id'.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('forcing_number', 'int', '0.1',
                "Identifies possible perturbatinos in forcing."),
            ('initialisation_method_number', 'int', '1.1',
                "Identifies which method of initialisation was used, if multiple methods used."),
            ('perturbation_number', 'int', '1.1',
                "Identifies different members of a perturbed physics ensemble."),
            ('realisation_number', 'int', '1.1',
                "Standard ensemble axis realisation number (usually an initial condition ensemble).")
        ]
    }


def drs_experiment():
    """An encoding of a drs experiment_id.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('axis_identifer', 'designing.axis_member', '0.1',
                "FIXME."),
            ('axis_type', 'designing.ensemble_types', '0.1',
                "Type of experimental family ensemble required."),
            ('family', 'str', '1.1',
                "Main experimental family.")
        ]
    }


def drs_frequency_types():
    """Set of allowed DRS frequency types.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("yr", "Yearly"),
            ("mon", "Monthly"),
            ("day", "Daily"),
            ("6hr", "Every six hours"),
            ("3hr", "Every three hours"),
            ("subhr", "Sampling frequency less than an hour"),
            ("monClim", "Climatological Monthly Mean"),
            ("fx", "Fixed Time independent")
        ]
    }


def drs_geographical_indicator():
    """Specifies geographical subsets described by bounding boxes or by named regions.
     One of spatial domain or bounding box must appear.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('bounding_box', 'str', '0.1',
                "DRS bounding box of the form 'latJHJJHHlonMZMMZZ' where H, HH, Z, ZZ are from {NS} {EW} respectively."),
            ('operator', 'drs.drs_geographical_operators', '0.1',
                "Spatial averagin applied to the geographical region."),
            ('spatial_domain', 'str', '0.1',
                "Geographical indicator (currently only 'global' is acceptable).")
        ]
    }


def drs_geographical_operators():
    """Set of permitted spatial averaging operator suffixes for drs spatial indicators (yyyy-zzzz).

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("zonalavg", "Data is zonally averaged"),
            ("lnd-zonalavg", "Data is zonally averaged over land in region"),
            ("ocn-zonalavg", "Data is zonally averaged over oceans in region"),
            ("areaavg", "Data is averaged over the area of the region"),
            ("lnd-areaavg", "Data is averaged over the land area of the region"),
            ("ocn-areaavg", "Data is averaged over the ocean area of the region")
        ]
    }


def drs_publication_dataset():
    """PLACEHOLDER for the real drs_publication_dataset.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
        ]
    }


def drs_realms():
    """Set of allowed DRS modelling realms.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("atmos", "Atmosphere"),
            ("ocean", "Ocean"),
            ("land", "Land"),
            ("landIce", "Land Ice"),
            ("seaIce", "Sea Ice"),
            ("aerosol", "Aerosol"),
            ("atmosChem", "Atmospheric Chemistry"),
            ("ocnBgchem", "Ocean Biogeochemistry")
        ]
    }


def drs_simulation_identifier():
    """That part of the DRS which identifies the response to the experiment: the simulation.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('institute', 'str', '1.1',
                "The institute responsible for this data entity."),
            ('model', 'str', '1.1',
                "A model identifier (sans blanks/periods and parenthesis)."),
            ('run_variant_id', 'drs.drs_ensemble_identifier', '1.1',
                "Also known as ensemble_identifier, unambiguously identifiers ensemble realisation information.")
        ]
    }


def drs_temporal_identifier():
    """Provides information about temporal subsetting and/or averaging.
    If only N1 is present, it a temporal instant,
    If N1-N2 are present with no suffix, it is a temporal subset,
    If N1-N2 with a suffix are present, then some sort of temporal averaging has been applied across
    the period.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('end', 'str', '0.1',
                "N2, required to indicate a period."),
            ('start', 'str', '1.1',
                "N1,  of the form yyyy[MM[dd[hh[mm[ss]]]]]."),
            ('suffix', 'drs.drs_time_suffixes', '0.1',
                "If present, used to indicate climatology or average.")
        ]
    }


def drs_time_suffixes():
    """Set of permitted time averaging suffixes for drs temporal identifiers.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("avg", "Indicates data is a single average of DRS frequency data across temporal period N1-N2"),
            ("clim", "Indicates data is climatological average data at the DRS frequency from the period N1-N2")
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: time.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def calendar():
    """Describes the calendar required/used in an experiment/simulation.
    This class is based on the calendar attributes and properties found in the
    CF netCDF conventions.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{}', ('standard_name',)),
        'properties': [
            ('description', 'str', '0.1',
                "Extra information about the calendar."),
            ('month_lengths', 'int', '0.N',
                "Used for special calendars to provide month lengths."),
            ('name', 'str', '0.1',
                "Can be used to name a special calendar type."),
            ('standard_name', 'time.calendar_types', '1.1',
                "Type of calendar used.")
        ]
    }


def calendar_types():
    """CF calendar types as defined in CF 1.6.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("gregorian", "Mixed Gregorian/Julian calendar as defined by Udunits"),
            ("standard", "Synonym for gregorian: Mixed Gregorian/Julian calendar as defined by Udunits"),
            ("proleptic_gregorian", "A Gregorian calendar extended to dates before 1582-10-15. That is, a year is a leap year if either (i) it is divisible by 4 but not by 100 or (ii) it is divisible by 400."),
            ("noleap", "Gregorian calendar without leap years, i.e., all years are 365 days long."),
            ("365_day", "Synonym for noleap:Gregorian calendar without leap years, i.e., all years are 365 days long."),
            ("all_leap", "Gregorian calendar with every year being a leap year, i.e., all years are 366 days long."),
            ("366_day", "Synonym for all_leap:Gregorian calendar with every year being a leap year, i.e., all years are 366 days long."),
            ("360_day", "All years are 360 days divided into 30 day months."),
            ("julian", "Julian Calendar"),
            ("none", "Perpetual time axis")
        ]
    }


def date_time():
    """A date or time. Either in simulation time with the simulation
    calendar, or with reference to a simulation start, in which
    case the datetime is an interval after the start date.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{}(offset {})', ('value', 'offset')),
        'properties': [
            ('is_offset', 'bool', '0.1',
                "Date is offset from start of an integration."),
            ('value', 'str', '1.1',
                "Date or time - some of (from left to right): yyyy-mm-dd:hh:mm:ss.")
        ]
    }


def datetime_set():
    """A set of times. This is an abstract class which is specialised into
    a periodic or aperiodic (irregular) list.  Note that we assume either a
    periodic set of dates which can be date and/or time, or an irregular set
    which can only be dates.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': True,
        'alternatives': [
            'time.irregular_dateset',
            'time.regular_timeset'
        ],
        'properties': [
            ('length', 'int', '1.1',
                "Number of times in set.")
        ]
    }


def irregular_dateset():
    """A set of irregularly spaced times, provided as a comma separated string of yyyy-mm-dd in
     the appropriate calendar.

    """
    return {
        'type': 'class',
        'base': 'time.datetime_set',
        'is_abstract': False,
        'properties': [
            ('date_set', 'str', '1.1',
                "Set of dates, comma separated yyyy-mm-dd.")
        ]
    }


def period_date_types():
    """A period date type vocabulary (used by time_period).

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("unused", "Date is not used"),
            ("date is start", "The date defines the start of the period"),
            ("date is end", "The date is the end of the period")
        ]
    }


def regular_timeset():
    """A regularly spaced set of times.

    """
    return {
        'type': 'class',
        'base': 'time.datetime_set',
        'is_abstract': False,
        'pstr': ('{} times from {} at {} intervals', ('length', 'start_date', 'increment')),
        'properties': [
            ('increment', 'time.time_period', '1.1',
                "Interval between members of set."),
            ('length', 'int', '1.1',
                "Number of times in set."),
            ('start_date', 'time.date_time', '1.1',
                "Beginning of time set.")
        ]
    }


def time_period():
    """A time period/interval (aka temporal extent).

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'pstr': ('{} {}', ('length', 'units')),
        'properties': [
            ('calendar', 'time.calendar', '0.1',
                "Calendar, default is standard aka gregorian."),
            ('date', 'time.date_time', '0.1',
                "Optional start/end date of time period."),
            ('date_type', 'time.period_date_types', '1.1',
                "Describes how the date is used to define the period."),
            ('length', 'str', '1.1',
                "Duration of the time period."),
            ('units', 'time.time_units', '1.1',
                "Appropriate time units.")
        ]
    }


def time_units():
    """Appropriate Time units for experiment durations in NWP and Climate Modelling.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("years", "Years in current calendar"),
            ("months", "Months in the current calendar"),
            ("days", "86400 seconds"),
            ("seconds", "Standard metric seconds")
        ]
    }



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land ice shelves'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'TODO',
    'properties': [
        ('adaptive_grid', 'bool', '1.1',
             'Is an adative grid being used?'),    
        ('resolution_limit', 'float', '0.1',
             'If an adative grid is being used, what is the limit of the resolution (in metres)'),    
        ('basal_melting', 'str', '1.1',
            'Describe the land ice shelf basal melting scheme'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Mass Balance
# --------------------------------------------------------------------
DETAILS['mass_balance'] = {
    'description': 'TODO',
    'properties': [
        ('calculated_in_atmosphere', 'bool', '1.1',
             'Is the ice shelf mass balance calculated in the atmosphere component?'),        
        ('ablation_calculation', 'ENUM:ablation_calculation_types', '1.1',
             'If mass balance is calculated in land ice model, the type of scheme used to calcualte ablation in the ice shelf mass balance'),
        ('downscaling_technique', 'str', '1.1',
             'If mass balance is calculated in land ice model, describe how the atmospheric variables are used in the mass balance calculations'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Dynamics
# --------------------------------------------------------------------
DETAILS['dynamics'] = {
    'description': 'TODO',
    'properties': [
        ('coupling_wth_ocean', 'str', '1.1',
             'Describe the coupling method between the land ice shelf and ocean'),
    ],
    'detail_sets': [
        'model_numerics',
    ]
}

DETAILS['dynamics:model_numerics'] = {
    'description': 'TODO',
    'properties': [
        ('timestep', 'int', '1.1',
             'Timestep (in seconds) of the land ice shelf scheme',),        
        ('uses_ocean_timestep', 'bool', '1.1',
             'Is the timestep the same as land surface timestep?'),
        ('uses_land_surface_timestep', 'bool', '1.1',
             'Is the timestep the same as land surface timestep?'),
        ('approximation', 'ENUM:approximation_types', '1.1',
            'Approximation type used in modelling ice shelf dynamics'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Snow treatment
# --------------------------------------------------------------------
DETAILS['snow_treatment'] = {
    'description': 'TODO',
    'properties': [
        ('calculated_in_atmosphere', 'bool', '1.1',
             'Is the ice shelf snow calculated in the atmosphere/land surface components?'),
        ('subgrid_hipsometry', 'str', '0.1',
             'Describe any subgrid-scale hipsometry in land ice shelf snow'),
        ('number_of_snow_layers', 'int', '0.1',
             'If mass balance is calculated in land ice model, how many layers in the ice shelf snow model?'),
        ('prognostic_variables', 'ENUM:snow_prognostic_variables', '0.N',
             'If mass balance is calculated in land ice model, list the prognostic variables the ice shelf snow model'),
        ('processes', 'ENUM:ice_shelf_snow_processes', '0.N',
             'If mass balance is calculated in land ice model, describe processes describing snow on ice shelves'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['bottom_types'] = {
    'description': 'TODO',
    'is_open': True,
    'members': [
        ('basal melting', None),
    ]
}

ENUMERATIONS['ablation_calculation_types'] = {
    'description': 'Type of scheme used to model ice shelf ablation',
    'is_open': True,
    'members': [
        ('Energy balance model', None),
        ('PDD', 'positive degree days'),
    ]
}

ENUMERATIONS['approximation_types'] = {
    'description': 'Approximation type used in modelling ice shelf dynamics',
    'is_open': True,
    'members': [
        ('SIA', None),
        ('SAA', None),
        ('full stokes', None),
        ]
    }

ENUMERATIONS['ice_shelf_snow_methods'] = {
    'description': 'Treatment of ice shelf snow',
    'is_open': True,
    'members': [
        ('same as snow in land surface', None),
        ('different from snow in land surface', None),
    ]
}

ENUMERATIONS['snow_prognostic_variables'] = {
    'description': 'Prognostic variables the ice shelf snow model',
    'is_open': True,
    'members': [
        ('snow albedo', None),
        ('snow density', None),
        ('snow water equivalent', None),
        ('snow heat content', None),
        ('snow temperature', None),
        ('snow liquid water content', None),
    ]
}

ENUMERATIONS['ice_shelf_snow_processes'] = {
    'description': 'Processes describing snow on ice shelves',
    'is_open': True,
    'members': [
        ('snow melting', None),
        ('snow refreezing', None),
        ('blowing snow', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: software_model_component.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.software.modelcomponent document.


"""
from collections import namedtuple

from pyesdoc._extensions.cim.v1.software_model_component_type_map import METAFOR_ESDOC_MODEL_COMPONENT_MAP
from pyesdoc._extensions.cim.v1.software_model_component_property_1 import extend as extend_property
from pyesdoc._extensions.cim.v1.software_model_component_property_2 import set_standard_properties
from pyesdoc._extensions.cim.v1.software_model_component_property_2 import set_scientific_properties
from pyesdoc._extensions.cim.v1.software_model_component_property_2 import set_qc_properties



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_component_hierarchy,
        _set_component_property_sets,
        _set_component_property_trees,
        _set_component_meta_info,
        _set_component_type_info
        )


# Mappings between component type & component type display name.
_COMPONENT_TYPE_DISPLAY_NAMES = \
    {k: v.split(">")[-1].strip() for k, v in iter(METAFOR_ESDOC_MODEL_COMPONENT_MAP.items())}


# Component extension context information.
_ComponentContextInfo = \
    namedtuple('ComponentContextInfo', ['c', 'ext', 'parent', 'ancestors'])


class _ComponentExtensionInfo(object):
    """Component extension properties.

    """
    pass


def _get_sort_key(item):
    """Returns a sort key.

    """
    return item.ext.full_display_name


def _extend_component_00(ctx):
    """Pre component extension tasks.

    """
    # Remove sub-components with invalid type names.
    for sc in ctx.c.sub_components:
        if sc.type.lower() == 'other' and \
           sc.short_name.replace(" ", "").lower() in \
                ['new', 'surfexoceansurface']:
            ctx.c.sub_components.remove(sc)

    # Rename component short name.
    if ctx.c.short_name == 'Ocean straits and horizontal diffusion':
        ctx.c.short_name = 'Ocean Straits'

    # If component type was specified as other then derive it from short name.
    if ctx.c.type.lower() == 'other':
        ctx.c.type = ctx.c.short_name.replace(" ", "")


def _extend_component_01(ctx):
    """Initializes component extension properties.

    """
    if not hasattr(ctx.c, "ext"):
        ctx.c.ext = _ComponentExtensionInfo()
    if not hasattr(ctx.c.ext, "full_display_name"):
        ctx.c.ext.full_display_name = unicode()
    if not hasattr(ctx.c.ext, "long_display_name"):
        ctx.c.ext.long_display_name = unicode()
    if not hasattr(ctx.c.ext, "short_display_name"):
        ctx.c.ext.short_display_name = unicode()
    if not hasattr(ctx.c.ext, "type_display_name"):
        ctx.c.ext.type_display_name = unicode()

    ctx.c.ext.ancestors = ctx.ancestors
    ctx.c.ext.component_tree = []
    ctx.c.ext.depth = len(ctx.ancestors)
    ctx.c.ext.parent = ctx.parent
    ctx.c.ext.properties = []
    ctx.c.ext.qc_properties = []
    ctx.c.ext.qc_property_tree = []
    ctx.c.ext.displayable_qc_properties = []
    ctx.c.ext.scientific_properties = []
    ctx.c.ext.scientific_property_tree = []
    ctx.c.ext.displayable_scientific_properties = []
    ctx.c.ext.standard_properties = []
    ctx.c.ext.standard_property_tree = []
    ctx.c.ext.displayable_standard_properties = []


def _extend_component_02(ctx):
    """Append component to tree of each direct ancestor.

    """
    for a in ctx.c.ext.ancestors:
        a.ext.component_tree.append(ctx.c)


def _extend_component_03(ctx):
    """Sets component display names.

    """
    # Type display name.
    if ctx.c.type in _COMPONENT_TYPE_DISPLAY_NAMES:
        ctx.c.ext.type_display_name = _COMPONENT_TYPE_DISPLAY_NAMES[ctx.c.type]
    else:
        ctx.c.ext.type_display_name = ctx.c.type

    # Short name.
    if not ctx.c.ext.parent:
        ctx.c.ext.short_display_name = ctx.c.short_name
    elif ctx.c.type in _COMPONENT_TYPE_DISPLAY_NAMES:
        ctx.c.ext.short_display_name = _COMPONENT_TYPE_DISPLAY_NAMES[ctx.c.type]
    else:
        ctx.c.ext.short_display_name = ctx.c.type

    # Long name.
    ctx.c.ext.long_display_name = unicode()
    if ctx.c.ext.depth > 1:
        ctx.c.ext.long_display_name += ctx.c.ext.parent.ext.long_display_name
        ctx.c.ext.long_display_name += u" > "
    ctx.c.ext.long_display_name += ctx.c.ext.short_display_name

    # Full name.
    if ctx.c.ext.depth > 0:
        ctx.c.ext.full_display_name += ctx.c.ext.parent.ext.full_display_name
        ctx.c.ext.full_display_name += u" > "
        ctx.c.ext.full_display_name += ctx.c.ext.short_display_name


def _extend_component_04(ctx):
    """Recursively extend child components.

    """
    for c in ctx.c.sub_components:
        _extend_component(c, c, ctx.c, ctx.ancestors + [ctx.c])


def _extend_component_05(ctx):
    """Sort components.

    """
    ctx.c.sub_components = sorted(ctx.c.sub_components, key=_get_sort_key)
    ctx.c.ext.component_tree = sorted(ctx.c.ext.component_tree, key=_get_sort_key)


def _extend_component(c, ext, parent=None, ancestors=[]):
    """Extends a component.

    """
    ctx = _ComponentContextInfo(c, ext, parent, ancestors)
    _extend_component_00(ctx)
    _extend_component_01(ctx)
    _extend_component_02(ctx)
    _extend_component_03(ctx)
    _extend_component_04(ctx)
    _extend_component_05(ctx)


def _set_component_hierarchy(ctx):
    """Extends component hierarchy.

    """
    _extend_component(ctx.doc, ctx.ext)


def _set_component_property_sets(ctx):
    """Extends component property sets.

    """
    def set_properties(builder, c, p_list):
        builder(c)
        for p in p_list:
            extend_property(c, p)
            c.ext.properties.append(p)

    for c in [ctx.doc] + ctx.doc.ext.component_tree:
        set_properties(set_scientific_properties, c, c.ext.scientific_properties)
        set_properties(set_standard_properties, c, c.ext.standard_properties)
        set_properties(set_qc_properties, c, c.ext.qc_properties)


def _set_component_property_trees(ctx):
    """Extends component property tree.

    """
    def build_tree(p_list, p_tree):
        """Builds a flatted property tree.

        """
        for p in p_list:
            p_tree.append(p)
            build_tree(p.sub_properties, p_tree)


    def get_displayable(p_tree):
        """Returns only displayable properties within a tree.

        """
        return [p for p in p_tree if p.values and p.values[0].lower() != 'none']

    for c in [ctx.doc] + ctx.doc.ext.component_tree:
        # ... build trees
        build_tree(c.ext.scientific_properties, c.ext.scientific_property_tree)
        build_tree(c.ext.standard_properties, c.ext.standard_property_tree)
        build_tree(c.ext.qc_properties, c.ext.qc_property_tree)
        # ... sort trees
        c.ext.scientific_property_tree = sorted(c.ext.scientific_property_tree, key=_get_sort_key)
        c.ext.standard_property_tree = sorted(c.ext.standard_property_tree, key=_get_sort_key)
        c.ext.qc_property_tree = sorted(c.ext.qc_property_tree, key=_get_sort_key)
        # ... set displayable properties
        c.ext.displayable_scientific_properties = get_displayable(c.ext.scientific_property_tree)
        c.ext.displayable_standard_properties = get_displayable(c.ext.standard_property_tree)
        c.ext.displayable_qc_properties = get_displayable(c.ext.qc_property_tree)


def _set_component_meta_info(ctx):
    """Extends component meta info.

    """
    for c in ctx.doc.ext.component_tree:
        for field in ('project', 'source', 'type'):
            if not getattr(c.meta, field):
                setattr(c.meta, field, getattr(ctx.doc.meta, field))


def _set_component_type_info(ctx):
    """Extends component type info.

    """
    for c in [ctx.doc] + ctx.doc.ext.component_tree:
        if not c.type:
            c.type = c.meta.type
        if c.meta.type not in c.types:
            c.types = [c.meta.type] + c.types



# -*- coding: utf-8 -*-

"""
.. module:: utils.config.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Configuration utility functions.

.. moduleauthor:: Mark Conway-Greenslade (formerly Morgan) <momipsl@ipsl.jussieu.fr>


"""
import os

from pyesdoc.utils.convert import json_file_to_namedtuple
from pyesdoc.utils.logger import log



# Users home directory.
_HOME = os.path.expanduser("~")

# Default configuration file path.
_CONFIG_FPATH = "pyesdoc.conf"

# Configuration data.
data = None


def _init_from_shell_config():
    """Attempts to initialize configuration from the shell.

    """
    # Find.
    dir_path = os.path.dirname(os.path.abspath(__file__))
    while dir_path != '/':
        config_path = os.path.join(dir_path, "ops/config/pyesdoc.conf")
        if os.path.exists(config_path):
            return config_path
        dir_path = os.path.dirname(dir_path)


def _init_from_local_config():
    """Attempts to initialize configuration from local file system.

    """
    # Find.
    dir_path = os.path.dirname(os.path.abspath(__file__))
    while dir_path != '/':
        config_path = os.path.join(dir_path, "pyesdoc.conf")
        if os.path.exists(config_path):
            return config_path
        dir_path = os.path.dirname(dir_path)


def _init_from_user_home():
    """Attempts to initialize configuration from a file foud on user's home directory.

    """
    config_path = os.path.join(_HOME, ".pyesdoc")
    if os.path.exists(config_path):
        return config_path


def init():
    """Initializes configuration.

    """
    global data

    # Escape if config data already initialized.
    if data is not None:
        return

    # Attempt to load config.
    config_path = _init_from_local_config() or \
                  _init_from_user_home() or \
                  _init_from_shell_config()

    if not config_path:
        raise RuntimeError("""
            WARNING :: pyesdoc configuration file not found.
            If running pyesdoc within ES-DOC shell the config file location should be:
                SHELL-HOME/ops/config/pyesdoc.conf.
            If running pyesdoc in standalone mode the config file location should be:
                $HOME/.pyesdoc
            Otherwise:
                1.  Download the following default configuration file:
                    http://bit.ly/2g3NUmQ
                2.  Copy downloaded file to $HOME/.pyesdoc
            If this problem persists them contact es-doc-support@list.woc.noaa.gov
        """)

    log("Loading pyesdoc config from: {}".format(config_path))
    data = json_file_to_namedtuple(config_path)



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.drq.content.wrapper.py
   :copyright: Copyright "February 27, 2016", IPSL
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps dreq.xml.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
class Wrapper(object):
    """Wraps dreq.xml.

    """
    def __init__(self, sections):
        """Instance constructor.

        :param list sections: Collection of content sections.

        """
        self._sections = sorted(sections, key=lambda i: i._TABLE.label.lower())
        for section in sections:
            setattr(self, str(section), section)


    def __repr__(self):
        """Instance representation.

        """
        return unicode(self._sections)


    def __len__(self):
        """Returns number of sections in managed collection.

        """
        return len(self._sections)


    def __contains__(self, key):
        """Returns flag indicating whether a certain section exists or not.

        """
        if key in self._sections:
            return True
        return self[key] is not None


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(self._sections)


    def __getattr__(self, key):
        """Returns a section.

        """
        return self._get_section(key)


    def __getitem__(self, key):
        """Returns a section.

        """
        return self._get_section(key)


    def _get_section(self, key):
        """Returns a section.

        """
        if isinstance(key, (slice, int)):
            return self._sections[key]
        key = unicode(key).strip().lower()
        for section in self._sections:
            if section._TABLE.label.lower() == key or \
               section._TABLE.label_drq.lower() == key:
                return section



# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_designing_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.designing package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_activity_package as activity



class AxisMember(object):
    """A concrete class within the cim v2 type system.

    PLACEHOLDER for the real axis_member.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(AxisMember, self).__init__()



class NumericalExperiment(activity.Activity):
    """A concrete class within the cim v2 type system.

    Defines a numerical experiment.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(NumericalExperiment, self).__init__()

        self.governing_mips = []                          # designing.Project (0.N)
        self.related_experiments = []                     # designing.NumericalExperiment (0.N)
        self.related_mips = []                            # designing.Project (0.N)
        self.required_period = None                       # designing.TemporalConstraint (1.1)
        self.requirements = []                            # designing.NumericalRequirement (0.N)
        self.tier = None                                  # int (0.1)


class NumericalRequirement(activity.Activity):
    """A concrete class within the cim v2 type system.

    A numerical requirement associated with a numerical experiment.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(NumericalRequirement, self).__init__()

        self.additional_requirements = []                 # designing.NumericalRequirement (0.N)
        self.is_conformance_info_required = None          # bool (1.1)
        self.is_conformance_requested = None              # bool (1.1)
        self.scope = None                                 # designing.NumericalRequirementScope (0.1)


class Project(activity.Activity):
    """A concrete class within the cim v2 type system.

    Describes a scientific project.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Project, self).__init__()

        self.governed_experiments = []                    # designing.NumericalExperiment (0.N)
        self.homepage = None                              # unicode (0.1)
        self.objectives = []                              # unicode (0.N)
        self.previous_projects = []                       # designing.Project (0.N)
        self.required_experiments = []                    # designing.NumericalExperiment (0.N)
        self.sub_projects = []                            # designing.Project (0.N)


class SimulationPlan(activity.Activity):
    """A concrete class within the cim v2 type system.

    Describes a simulation that needs to be run.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(SimulationPlan, self).__init__()

        self.expected_model = None                        # science.Model (1.1)
        self.expected_performance_sypd = None             # float (0.1)
        self.expected_platform = None                     # platform.Machine (0.1)
        self.will_support_experiments = []                # designing.NumericalExperiment (1.N)


class DomainRequirements(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    Properties of the domain which needs to be simulated, extent and/or resolution.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DomainRequirements, self).__init__()

        self.required_extent = None                       # science.Topic (0.1)
        self.required_resolution = None                   # science.Topic (0.1)


class EnsembleRequirement(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    Defines an experiment ensemble.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(EnsembleRequirement, self).__init__()

        self.ensemble_member = []                         # designing.NumericalRequirement (0.N)
        self.ensemble_type = None                         # designing.EnsembleTypes (1.1)
        self.minimum_size = None                          # int (1.1)


class ForcingConstraint(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    Identifies a model forcing constraint.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(ForcingConstraint, self).__init__()

        self.additional_constraint = None                 # unicode (0.1)
        self.category = None                              # unicode (0.1)
        self.code = None                                  # unicode (0.1)
        self.data_link = None                             # data.Dataset (0.1)
        self.forcing_type = None                          # designing.ForcingTypes (1.1)
        self.group = None                                 # unicode (0.1)
        self.origin = None                                # shared.Citation (0.1)


class InitialisationRequirement(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    A requirement on how a particular simulation should be initialised.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(InitialisationRequirement, self).__init__()

        self.branch_time_in_initialisation_source = None  # time.DateTime (0.1)
        self.initialise_from_data = None                  # data.Dataset (0.1)
        self.initialise_from_experiment = None            # designing.NumericalExperiment (0.1)


class MultiEnsemble(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    In the case of multiple ensemble axes, defines how they
    are set up and ordered.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(MultiEnsemble, self).__init__()

        self.ensemble_axis = []                           # designing.EnsembleRequirement (1.N)


class OutputRequirement(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    Provides details of what output is required and when from an experiment.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(OutputRequirement, self).__init__()

        self.formal_data_request = None                   # shared.OnlineResource (0.1)


class StartDateEnsemble(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    Defines an experiment ensemble with multiple start dates.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(StartDateEnsemble, self).__init__()

        self.ensemble_members = None                      # time.DatetimeSet (1.1)


class TemporalConstraint(NumericalRequirement):
    """A concrete class within the cim v2 type system.

    A spatio-temporal constraint on a numerical experiment.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(TemporalConstraint, self).__init__()

        self.required_calendar = None                     # time.Calendar (0.1)
        self.required_duration = None                     # time.TimePeriod (0.1)
        self.start_date = None                            # time.DateTime (0.1)
        self.start_flexibility = None                     # time.TimePeriod (0.1)


class EnsembleTypes(object):
    """An enumeration within the cim v2 type system.

    Defines the various axes along which one can set up an ensemble, whether
     as an experiment designer, or in designing a 'response' ensemble around an
     experiment.
    """
    is_open = False
    members = [
        "Perturbed Physics",
        "Initialisation Method",
        "Realization",
        "Start Date",
        "Forced",
        "Resolution",
        "Driven"
        ]
    descriptions = [
        "Members differ in some aspects of their physics.",
        "Members differ in how they are initialised.",
        "Members initialised to sample possible initial conditions.",
        "Members initialised at different starting dates.",
        "Members have differing forcing data.",
        "Members are/should-be run at different resolutions.",
        "Members are/should-be driven by different models."
        ]


class ExperimentalRelationships(object):
    """An enumeration within the cim v2 type system.

    Defines the canonical set of experimental relationships.
    """
    is_open = True
    members = [
        "is_constrained_by",
        "is_constrainer_of",
        "is_perturbation_from",
        "is_control_for",
        "is_initialized_by",
        "is_initializer_of",
        "is_sibling_of"
        ]
    descriptions = [
        "The experiment that provides constraint(s) for the target experiment (e.g SST forcing).",
        "The set of experiments constrained by the experiment.",
        "The experiment that provides a control for the target experiment.",
        "The set of experiments that use the experiment as a control.",
        "The experiment providing initialization for the experiment.",
        "The set of experiments initialized by the experiment.",
        "Part of a family (e.g. experiments where solar forcing is either increased or reduced)."
        ]


class ForcingTypes(object):
    """An enumeration within the cim v2 type system.

    Defines the possible set of forcing types for a forcing constraint.
    """
    is_open = False
    members = [
        "historical",
        "idealised",
        "scenario",
        "driven"
        ]
    descriptions = [
        "Best estimates of actual state (included synthesized)",
        "Simplified and/or exemplar, e.g. 1%C02",
        "Intended to represent a possible future, e.g. RCP4.5",
        "Driven from another simulation"
        ]


class NumericalRequirementScope(object):
    """An enumeration within the cim v2 type system.

    The scope to which a numerical requirement may or may not apply.
    """
    is_open = False
    members = [
        "mip-era",
        "mip-group",
        "mip",
        "experiment"
        ]
    descriptions = [
        "MIP era wide e.g. 'concentration of pre-industrial CO2' & 'Impose AMIP SSTs'",
        "Shared across a group of MIPs e.g. aerosol forcing in GeoMIP and AerChemMIP.",
        "Shared within a single MIP e.g. spin-up protocol for land surface experiments in LUMIP.",
        "Applies to a single experiment e.g. CFMIP 'zonally uniform SST plus 4K'"
        ]





"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Aerosol model'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level aerosol model properties',
    'properties': [
        ('processes', 'ENUM:model_processes', '1.N',
            'Processes included in the Aerosol model.'),
        ('coupling', 'ENUM:coupling', '0.N',
            'Other model components coupled to the Aerosol model'),
        ('gas_phase_precursors', 'ENUM:gas_phase_precursors', '1.N',
            'List of gas phase aerosol precursors.'),
        ('scheme_type', 'ENUM:scheme_types', '1.N',
             'Type(s) of aerosol scheme used by the aerosols model (potentially multiple: some species may be covered by one type of aerosol scheme and other species covered by another type).'),
        ('bulk_scheme_species', 'ENUM:species', '1.N',
             'List of species covered by the bulk scheme.'),
        ('model_scheme_framework', 'ENUM:scheme_frameworks', '1.1',
             'Aerosols classification framework used in the modal aerosol scheme.'),
        ('model_scheme_species', 'ENUM:species', '1.N',
             'List of species covered by the modal scheme.'),
        ('bin_scheme_framework', 'ENUM:scheme_frameworks', '1.1',
             'Aerosols classification framework used in the bin aerosol scheme.'),
        ('bin_scheme_species', 'ENUM:species', '1.N',
             'List of species covered by the bin scheme.'),
        ('classification_framework', 'ENUM:scheme_frameworks', '0.1',
             'Aerosols classification framework.'),
        ('classification_framework_characterisitics', 'str', '0.1',
             'Characteristics of the aerosol scheme used.'),
        ('classification_framework_species', 'ENUM:species', '0.1',
             'List of species covered by the scheme.'),
    ]
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_processes'] = {
    'description': 'Processes included in the Aerosol model.',
    'is_open': False,
    'members': [
        ('Dry deposition', None),
        ('Sedimentation', None),
        ('Wet deposition (impaction scavenging)', None),
        ('Wet deposition (nucleation scavenging)', None),
        ('Coagulation', None),
        ('Oxidation (gas phase)', None),
        ('Oxidation (in cloud)', None),
        ('Condensation', None),
        ('Ageing', None),
        ('Advection (horizontal)', None),
        ('Advection (vertical)', None),
        ('Heterogeneous chemistry', None),
    ]
}

ENUMERATIONS['coupling'] = {
    'description':'Other model components coupled to the Aerosol model.',
    'is_open': True,
    'members':[
        ('Radiation', None),
        ('Land surface', None),
        ('Heterogeneous chemistry', None),
        ('Clouds', None),
    ]
}

ENUMERATIONS['gas_phase_precursors'] = {
    'description': 'List of gas phase aerosol precursors.',
    'is_open': True,
    'members':[
        ('DMS', None),
        ('SO2', None),
        ('Ammonia', None),
        ('Iodene', None),
        ('Terpene', None),
        ('Isoprene', None),
    ]
}

ENUMERATIONS['scheme_types'] = {
    'description': 'Type(s) of aerosol scheme used by the aerosols model (potentially multiple: some species may be covered by one type of aerosol scheme and other species covered by another type).',
    'is_open': True,
    'members':[
        ('Bulk', None),
        ('Modal', None),
        ('Bin', None),
    ]
}

ENUMERATIONS['species'] = {
    'description':'List of species.',
    'is_open': True,
    'members':[
        ('Sulphate', None),
        ('Nitrate', None),
        ('Sea salt', None),
        ('Dust', None),
        ('Ice', None),
        ('Organic', None),
        ('Black carbon / soot', None),
        ('SOA (secondary organic aerosols)', None),
        ('POM (particulate organic matter)', None),
        ('Polar stratospheric ice', None),
        ('NAT (Nitric acid trihydrate)', None),
        ('NAD (Nitric acid dihydrate)', None),
        ('STS (supercooled ternary solution aerosol particule)', None),
    ]
}

ENUMERATIONS['scheme_frameworks'] = {
    'description': 'Aerosols classification framework',
    'is_open': True,
    'members':[
        ('Single moment', None),
        ('Microphysical', None),
    ]
}




"""A realm key-properties specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Key properties of the atmospheric chemistry'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS = OrderedDict()

DETAILS['toplevel'] = {
    'description': 'Top level key properties in atmospheric chemistry model',
    'properties': [
        ('chemistry_scheme_scope', 'ENUM:chemistry_scheme_scopes', '1.N',
            'Atmospheric domains covered by the atmospheric chemistry model'),
        ('basic_approximations', 'str', '1.1',
            'Basic approximations made in the atmospheric chemistry model',),
        ('prognostic_variables_form', 'ENUM:prognostic_vars_types', '1.N',
            'Form of prognostic variables in the atmospheric chemistry component.'),
        ('number_of_tracers', 'int', '1.1',
            'Number of tracers in the atmospheric chemistry model'),
        ('family_approach', 'bool', '1.1',
            'atmospheric chemistry calculations generalized into families of species?'),
        ('coupling_with_chemical_reactivity', 'bool', '1.1',
            'Atmospheric chemistry transport scheme turbulence is couple with chemical reactivity?'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: timestepping framework
# --------------------------------------------------------------------
DETAILS['timestep_framework'] = {
    'description': 'Physical properties of seawater in ocean',
    'properties' : [
        ('method', 'ENUM:timestepping_methods', '1.1',
            'Mathematical method deployed to solve teh evolution of a given variable'),
        ('split_operator_advection_timestep', 'int', '0.1',
            'Timestep for chemical species advection (in seconds)'),
        ('split_operator_physical_timestep', 'int', '0.1',
            'Timestep for physics and chemistry (in seconds).'),
        ('split_operator_alternate_order', 'bool', '1.1',
            'TO DO'),
        ('integrated_timestep', 'int', '1.1',
            'Timestep for the atmospheric chemistry model (in seconds)'),
        ('integrated_scheme_type', 'ENUM:integrated_scheme_types', '1.1',
            'Specify the type of timestep scheme'),
    ],
    'detail_sets': [
        'split_operator_order',
    ]
}

DETAILS['timestep_framework:split_operator_order'] = {
    'description': '',
    'properties' : [
        ('turbulence', 'int', '0.1',
            'Call order for  scheme'),
        ('convection', 'int', '0.1',
            'Call order for convection scheme'),
        ('precipitation', 'int', '0.1',
            'Call order for precipitation scheme'),
        ('emissions', 'int', '0.1',
            'Call order for emissions scheme'),
        ('deposition', 'int', '0.1',
            'Call order for deposition scheme'),
        ('gas_phase_chemistry', 'int', '0.1',
            'Call order for gas phase chemistry scheme'),
        ('heterogeneous_phase_chemistry', 'int', '0.1',
            'Call order for heterogeneous phase chemistry scheme'),
        ('photo_chemistry', 'int', '0.1',
            'Call order for photo chemistry scheme'),
        ('aerosols', 'int', '0.1',
            'Call order for aerosols scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: Meteorological forcings
# --------------------------------------------------------------------
DETAILS['meteorological_forcings'] = {
    'description': '',
    'properties' : [
        ('variables_3D', 'str', '0.1',
            'Three dimensionsal forcing variables, e.g. U, V, W, T, Q, P, conventive mass flux'),
        ('variables_2D', 'str', '0.1',
            'Two dimensionsal forcing variables, e.g. land-sea mask definition'),
        ('frequency', 'int', '1.1',
            'Frequency with which meteological forcings are applied (in seconds).'),
        ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: RESOLUTION: The resolution of the grid.
# --------------------------------------------------------------------
DETAILS['resolution'] = {
    'description': 'Resolution in the atmospheric chemistry grid',
    'properties': [
        ('name', 'str', '1.1',
             "This is a string usually used by the modelling group to describe the resolution of this grid, e.g. ORCA025, N512L180, T512L70 etc."),
        ('canonical_horizontal_resolution', 'str', '0.1',
             "Expression quoted for gross comparisons of resolution, eg. 50km or 0.1 degrees etc."),
        ('number_of_horizontal_gridpoints', 'int', '0.1',
             "Total number of horizontal (XY) points (or degrees of freedom) on computational grid."),
        ('number_of_vertical_levels', 'int', '0.1',
             "Number of vertical levels resolved on computational grid."),
        ('is_adaptive_grid', 'bool', '0.1',
             "Default is False. Set true if grid resolution changes during execution."),
        ],
    }

# --------------------------------------------------------------------
# SUB-PROCESS: TUNING APPLIED: Any tuning used to optimise the parameters
# --------------------------------------------------------------------
DETAILS['tuning_applied'] = {
    'description': 'Tuning methodology for atmospheric chemistry component',
    'properties': [
        ('description', 'str', '1.1',
             "General overview description of tuning: explain and motivate the main targets and metrics retained. &"
             "Document the relative weight given to climate performance metrics versus process oriented metrics, &"
             "and on the possible conflicts with parameterization level tuning. In particular describe any struggle &"
             "with a parameter value that required pushing it to its limits to solve a particular model deficiency."),
        ('global_mean_metrics_used', 'str', '0.N',
             "List set of metrics of the global mean state used in tuning model/component"),
        ('regional_metrics_used', 'str', '0.N',
             "List of regional metrics of mean state used in tuning model/component"),
        ('trend_metrics_used', 'str', '0.N',
             "List observed trend metrics used in tuning model/component"),
        ]
    }

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS = OrderedDict()

ENUMERATIONS['chemistry_scheme_scopes'] = {
    'description': 'Atmospheric domains covered by the atmospheric chemistry model',
    'is_open': True,
    'members': [
        ('troposhere', None),
        ('stratosphere', None),
        ('mesosphere', None),
        ('mesosphere', None),
        ('whole atmosphere', None),
    ]
}

ENUMERATIONS['prognostic_vars_types'] = {
    'description': 'Form of prognostic variables in the atmospheric chemistry component.',
    'is_open': True,
    'members': [
        ('3D mass/mixing ratio for gas', None),
    ]
}

ENUMERATIONS['timestepping_methods'] = {
    'description': 'Mathematical method deployed to solve teh evolution of a given variable',
    'is_open': True,
    'members': [
        ('Operator splitting', None),
        ('Integrated', None),
    ]
}
     
ENUMERATIONS['integrated_scheme_types'] = {
    'description': 'Specify the type of timestep scheme',
    'is_open': True,
    'members': [
        ('Explicit', None),
        ('Implicit', None),
        ('Semi-implicit', None),
        ('Semi-analytic', None),
        ('Impact solver', None),
        ('Back Euler', None),
        ('Newton Raphson', None),
        ('Rosenbrock', None),
    ]
}



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmospheric chemistry transport'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Gas phase chemistry attributes',
    'properties': [
        ('species', 'ENUM:species_types', '0.N',
            'Species included in the gas phase chemistry scheme.'),
        ('number_of_bimolecular_reactions', 'int', '1.1',
             'The number of bi-molecular reactions in the gas phase chemistry scheme.'),
        ('number_of_termolecular_reactions', 'int', '1.1',
             'The number of ter-molecular reactions in the gas phase chemistry scheme.'),
        ('number_of_advected_species', 'int', '1.1',
             'The number of advected species in the gas phase chemistry scheme.'),
        ('number_of_steady_state_species', 'int', '1.1',
             'The number of steady state species in the gas phase chemistry scheme.'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Dry deposotion
# --------------------------------------------------------------------
DETAILS['dry_deposition'] = {
    'description': 'Dry deposition describes the dry processes by which gaseous species deposit themselves on solid surfaces thus decreasing their concentration in the air.',
    'properties': [
        ('interactive', 'bool', '1.1',
            'Is dry deposition interactive (as opposed to prescribed)?'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Wet deposotion
# --------------------------------------------------------------------
DETAILS['wet_deposition'] = {
    'description': 'Wet deposition describes the moist processes by which gaseous species deposit themselves on solid surfaces thus decreasing their concentration in the air.',
    'properties': [
        ('included', 'bool', '1.1',
            'Is wet deposition included?'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Oxidation
# --------------------------------------------------------------------
DETAILS['oxidation'] = {
    'description': 'Oxidation describes the loss of electrons or an increase in oxidation state by a molecule.',
    'properties': [
        ('included', 'bool', '1.1',
            'Is wet oxidation included?'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Washout-rainout
# --------------------------------------------------------------------
DETAILS['washout_rainout'] = {
    'description': 'Downward transport of soluble chemistry species in clouds.',
    'properties': [
        ('methods', 'ENUM:washout_methods', '0.N',
            'Methods of downward transport of soluble chemistry species in clouds.'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['washout_methods'] = {
    'description': 'Methods of downward transport of soluble chemistry species in clouds.',
    'is_open': False,
    'members': [
        ('In-cloud scavenging', None),
        ('Below-cloud scavenging', None),
        ]
    }

ENUMERATIONS['species_types'] = {
    'description':'Species included in the gas phase chemistry scheme.',
    'is_open': True,
    'members':[
        ('HOx', None),
        ('NOy', None),
        ('Ox', None),
        ('Cly', None),
        ('HSOx', None),
        ('Bry', None),
        ('VOCs', None),
        ('isoprene', None),
        ('H2O', None),
    ]
}



# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.ontologies.utils_graph.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Graph utilities.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.ontologies.type_info import CLASSES
from pyesdoc.ontologies.type_info import CLASS_PROPERTIES



class GraphParser(object):
    """Encapsulates object graph parsing.

    """
    def parse(self, container, path=''):
        """Performs a parse over a container raising events as it does so.

        """
        # Get properties.
        try:
            properties = CLASS_PROPERTIES[type(container)]
        except KeyError:
            raise ValueError("Container is an unsupported ontology class.")

        # Raise container parse event.
        self.on_container_parse(container, path)

        # Parse container properties.
        for p_name in properties:
            # Raise property parse event.
            p_value = getattr(container, p_name)
            p_path = "{}.{}".format(path, p_name)
            self.on_property_parse(container, p_path, p_name, p_value)

            # Parse list values.
            if isinstance(p_value, list):
                for idx, item in enumerate(p_value):
                    item_path = "{}.{}".format(p_path, idx)
                    self.on_list_item_parse(p_value, item_path, idx, item)
                    if isinstance(item, CLASSES):
                        self.parse(item, item_path)

            # Parse sub-containers.
            elif isinstance(p_value, CLASSES):
                self.parse(p_value, p_path)

        # Raise container parsed event.
        self.on_container_parsed(container)


    def on_container_parse(self, container, path):
        """Container parse event handler.

        """
        # Allow sub-classes to handle.
        pass


    def on_container_parsed(self, container):
        """Container parsed event handler.

        """
        # Allow sub-classes to handle.
        pass


    def on_property_parse(self, container, path, name, value):
        """Property parse event handler.

        """
        # Allow sub-classes to handle.
        pass


    def on_list_item_parse(self, container, path, idx, value):
        """List item parse event handler.

        """
        # Allow sub-classes to handle.
        pass



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotee Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmospheric chemistry transport'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level lateral physics properties',
    'properties': [
        ('scheme', 'ENUM:transport_scheme_types', '1.1',
            'Type of atmospheric chemistry transport scheme'),
        ('mass_conservation', 'ENUM:mass_conservation_methods', '1.N',
            'Method used to ensure mass conservation'),
        ('convection', 'ENUM:convection_types', '1.N',
            'Transport by convection'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Turbulence
# --------------------------------------------------------------------
DETAILS['turbulence'] = {
    'description': 'Properties of lateral physics for momentum in ocean',
    'properties': [
        ('dimensions', 'ENUM:turbulence_dimensions', '1.1',
            'Dimension of the turbulence scheme in chemistry transport'),
        ('coupling_with_chemical_reactivity', 'bool', '1.1',
             'Is the chemistry transport scheme turbulence coupled with chemical reactivity?'),
        
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['transport_scheme_types'] = {
    'description': 'Type of atmospheric chemistry transport scheme',
    'is_open': False,
    'members': [
        ('Eulerian', None),
        ('Semi-lagrangian', None),
        ('Eulerian and semi-lagrangian', None),
        ('Lagrangian', None),
        ]
    }

ENUMERATIONS['mass_conservation_methods'] = {
    'description':'Method used to ensure mass conservation',
    'is_open': True,
    'members':[
        ('Mass adjustment', None),
        ('Concentrations positivity', None),
        ('Gradients monotonicity', None),
    ]
}

ENUMERATIONS['convection_types'] = {
    'description':'Transport by convection',
    'is_open': True,
    'members':[
        ('Convective fluxes connected to tracers', None),
        ('Vertical velocities connected to tracers', None),
    ]
}

ENUMERATIONS['turbulence_dimensions'] = {
    'description':'TDimension of the turbulence scheme in chemistry transport',
    'is_open': True,
    'members':[
        ('2D', 'Two-dimensional'),
        ('3D', 'Three-dimensional'),
    ]
}



# -*- coding: utf-8 -*-
"""
.. module:: handlers_metafor_q.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes Metafor Questionnaire document processing handlers.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

# CIM v1 type keys.
_TYPE_KEY_MODEL_COMPONENT = 'cim.1.software.modelcomponent'

# Institute name overrides.
_INSTITUTE_OVERRIDES = {
    'DEPARTMENT OF ENERGY' : 'DOE',
}


def _set_document_type(doc):
    """Sets document type when unassigned.

    """
    if doc.type_key.lower() == _TYPE_KEY_MODEL_COMPONENT and not doc.types:
        if not doc.type:
            doc.type = _TYPE_KEY_MODEL_COMPONENT
        doc.types = [doc.type]


def _set_institute(doc):
    """Sets institute name.

    """
    if doc.meta.institute is None or not doc.meta.institute.strip():
        doc.meta.institute = u"--"
    elif doc.meta.institute.upper() in _INSTITUTE_OVERRIDES:
        doc.meta.institute = _INSTITUTE_OVERRIDES[doc.meta.institute.upper()]


# Set of document parsers.
PARSERS = (
    _set_document_type,
    _set_institute
    )


# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.decoder_for_software_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: A set of cim 1 decoders.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
from decoder_xml_utils import set_attributes
from decoder_for_activity_package import *
from decoder_for_data_package import *
from decoder_for_grids_package import *
from decoder_for_shared_package import *
import typeset



def decode_component(xml, nsmap):
    """Decodes an instance of the following type: component.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Component

    """
    decodings = [
        ('citations', True, decode_citation, 'child::cim:citation'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('language', False, decode_component_language, 'child::cim:componentLanguage'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('properties', True, decode_component_property, 'child::cim:componentProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:numericalProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:scientificProperties/cim:componentProperty'),
        ('release_date', False, 'datetime.datetime', 'child::cim:releaseDate'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('sub_components', True, decode_model_component, 'child::cim:childComponent/cim:modelComponent'),
        ('sub_components', True, decode_processor_component, 'child::cim:childComponent/cim:processorComponent'),
    ]

    return set_attributes(typeset.software.Component(), xml, nsmap, decodings)


def decode_component_language(xml, nsmap):
    """Decodes an instance of the following type: component language.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.ComponentLanguage

    """
    decodings = [
        ('name', False, 'unicode', 'child::cim:name'),
    ]

    return set_attributes(typeset.software.ComponentLanguage(), xml, nsmap, decodings)


def decode_component_language_property(xml, nsmap):
    """Decodes an instance of the following type: component language property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.ComponentLanguageProperty

    """
    decodings = [
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.software.ComponentLanguageProperty(), xml, nsmap, decodings)


def decode_component_property(xml, nsmap):
    """Decodes an instance of the following type: component property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.ComponentProperty

    """
    decodings = [
        ('citations', True, decode_citation, 'child::cim:citation'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('intent', False, 'unicode', 'self::cim:componentProperty/@intent'),
        ('is_represented', False, 'bool', 'self::cim:componentProperty/@represented'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('standard_names', True, 'unicode', 'child::cim:standardName/@value'),
        ('sub_properties', True, decode_component_property, 'child::cim:componentProperty'),
        ('units', False, 'unicode', 'child::cim:units/@value'),
        ('values', True, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.software.ComponentProperty(), xml, nsmap, decodings)


def decode_composition(xml, nsmap):
    """Decodes an instance of the following type: composition.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Composition

    """
    decodings = [
    ]

    return set_attributes(typeset.software.Composition(), xml, nsmap, decodings)


def decode_connection(xml, nsmap):
    """Decodes an instance of the following type: connection.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Connection

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('priming', False, decode_component_property, 'child::cim:priming/cim:priming/cim:componentProperty'),
        ('priming', False, decode_data_content, 'child::cim:priming/cim:priming/cim:dataContent'),
        ('priming', False, decode_data_object, 'child::cim:priming/cim:priming/cim:dataObject'),
        ('priming', False, decode_doc_reference, 'child::cim:priming/cim:reference'),
        ('priming', False, decode_model_component, 'child::cim:priming/cim:priming/cim:softwareComponent'),
        ('priming', False, decode_processor_component, 'child::cim:priming/cim:priming/cim:softwareComponent'),
        ('priming', False, decode_statistical_model_component, 'child::cim:priming/cim:priming/cim:softwareComponent'),
        ('properties', True, decode_connection_property, 'child::cim:connectionProperty'),
        ('sources', True, decode_connection_endpoint, 'child::cim:connectionSource'),
        ('spatial_regridding', True, decode_spatial_regridding, 'child::cim:spatialRegridding'),
        ('target', False, decode_connection_endpoint, 'child::cim:connectionTarget'),
        ('time_lag', False, 'unicode', 'child::cim:timeLag'),
        ('time_profile', False, decode_timing, 'child::cim:timeProfile'),
        ('time_transformation', False, decode_time_transformation, 'child::cim:timeTransformation'),
        ('transformers', True, decode_doc_reference, 'child::cim:transformer/cim:reference'),
        ('transformers', True, decode_processor_component, 'child::cim:transformer/cim:processorComponent'),
        ('type', False, 'unicode', 'child::cim:type/@value'),
    ]

    return set_attributes(typeset.software.Connection(), xml, nsmap, decodings)


def decode_connection_endpoint(xml, nsmap):
    """Decodes an instance of the following type: connection endpoint.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.ConnectionEndpoint

    """
    decodings = [
        ('data_source', False, decode_component_property, 'child::cim:dataSource/cim:dataSource/cim:componentProperty'),
        ('data_source', False, decode_data_content, 'child::cim:dataSource/cim:dataSource/cim:dataContent'),
        ('data_source', False, decode_data_object, 'child::cim:dataSource/cim:dataSource/cim:dataObject'),
        ('data_source', False, decode_doc_reference, 'child::cim:dataSource/cim:reference'),
        ('data_source', False, decode_model_component, 'child::cim:dataSource/cim:dataSource/cim:softwareComponent'),
        ('data_source', False, decode_processor_component, 'child::cim:dataSource/cim:dataSource/cim:softwareComponent'),
        ('data_source', False, decode_statistical_model_component, 'child::cim:dataSource/cim:dataSource/cim:softwareComponent'),
        ('instance_id', False, 'unicode', 'child::cim:instanceID'),
        ('properties', True, decode_connection_property, 'child::cim:connectionProperty'),
    ]

    return set_attributes(typeset.software.ConnectionEndpoint(), xml, nsmap, decodings)


def decode_connection_property(xml, nsmap):
    """Decodes an instance of the following type: connection property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.ConnectionProperty

    """
    decodings = [
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.software.ConnectionProperty(), xml, nsmap, decodings)


def decode_coupling(xml, nsmap):
    """Decodes an instance of the following type: coupling.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Coupling

    """
    decodings = [
        ('connections', True, decode_connection, 'child::cim:connection'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('is_fully_specified', False, 'bool', '@fullySpecified'),
        ('priming', False, decode_component_property, 'child::cim:priming/cim:priming/cim:componentProperty'),
        ('priming', False, decode_data_content, 'child::cim:priming/cim:priming/cim:dataContent'),
        ('priming', False, decode_data_object, 'child::cim:priming/cim:priming/cim:dataObject'),
        ('priming', False, decode_doc_reference, 'child::cim:priming/cim:reference'),
        ('priming', False, decode_model_component, 'child::cim:priming/cim:priming/cim:softwareComponent'),
        ('priming', False, decode_processor_component, 'child::cim:priming/cim:priming/cim:softwareComponent'),
        ('priming', False, decode_statistical_model_component, 'child::cim:priming/cim:priming/cim:softwareComponent'),
        ('properties', True, decode_coupling_property, 'child::cim:couplingProperty'),
        ('purpose', False, 'unicode', '@purpose'),
        ('sources', True, decode_coupling_endpoint, 'child::cim:couplingSource'),
        ('spatial_regriddings', True, decode_spatial_regridding, 'child::cim:spatialRegridding'),
        ('target', False, decode_coupling_endpoint, 'child::cim:couplingTarget'),
        ('time_lag', False, decode_time_lag, 'child::cim:timeLag'),
        ('time_profile', False, decode_timing, 'child::cim:timeProfile'),
        ('time_transformation', False, decode_time_transformation, 'child::cim:timeTransformation'),
        ('transformers', True, decode_doc_reference, 'child::cim:transformer/cim:reference'),
        ('transformers', True, decode_processor_component, 'child::cim:transformer/cim:processorComponent'),
        ('type', False, 'unicode', 'child::cim:type/@value'),
    ]

    return set_attributes(typeset.software.Coupling(), xml, nsmap, decodings)


def decode_coupling_endpoint(xml, nsmap):
    """Decodes an instance of the following type: coupling endpoint.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.CouplingEndpoint

    """
    decodings = [
        ('data_source', False, decode_component_property, 'child::cim:dataSource/cim:dataSource/cim:componentProperty'),
        ('data_source', False, decode_data_content, 'child::cim:dataSource/cim:dataSource/cim:dataContent'),
        ('data_source', False, decode_data_object, 'child::cim:dataSource/cim:dataSource/cim:dataObject'),
        ('data_source', False, decode_doc_reference, 'child::cim:dataSource/cim:reference'),
        ('data_source', False, decode_model_component, 'child::cim:dataSource/cim:dataSource/cim:softwareComponent'),
        ('data_source', False, decode_processor_component, 'child::cim:dataSource/cim:dataSource/cim:softwareComponent'),
        ('data_source', False, decode_statistical_model_component, 'child::cim:dataSource/cim:dataSource/cim:softwareComponent'),
        ('instance_id', False, 'unicode', 'child::cim:instanceID'),
        ('properties', True, decode_coupling_property, 'child::cim:couplingProperty'),
    ]

    return set_attributes(typeset.software.CouplingEndpoint(), xml, nsmap, decodings)


def decode_coupling_property(xml, nsmap):
    """Decodes an instance of the following type: coupling property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.CouplingProperty

    """
    decodings = [
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.software.CouplingProperty(), xml, nsmap, decodings)


def decode_deployment(xml, nsmap):
    """Decodes an instance of the following type: deployment.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Deployment

    """
    decodings = [
        ('deployment_date', False, 'datetime.datetime', 'child::cim:deploymentDate'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('executable_arguments', True, 'unicode', 'child::cim:executableArgument'),
        ('executable_name', False, 'unicode', 'child::cim:executableName'),
        ('parallelisation', False, decode_parallelisation, 'child::cim:parallelisation'),
        ('platform', False, decode_doc_reference, 'child::cim:platform/cim:reference'),
        ('platform', False, decode_platform, 'child::cim:platform/cim:platform'),
    ]

    return set_attributes(typeset.software.Deployment(), xml, nsmap, decodings)


def decode_entry_point(xml, nsmap):
    """Decodes an instance of the following type: entry point.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.EntryPoint

    """
    decodings = [
    ]

    return set_attributes(typeset.software.EntryPoint(), xml, nsmap, decodings)


def decode_model_component(xml, nsmap):
    """Decodes an instance of the following type: model component.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.ModelComponent

    """
    decodings = [
        ('citations', True, decode_citation, 'child::cim:citation'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('language', False, decode_component_language, 'child::cim:componentLanguage'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:modelComponent'),
        ('properties', True, decode_component_property, 'child::cim:componentProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:numericalProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:scientificProperties/cim:componentProperty'),
        ('release_date', False, 'datetime.datetime', 'child::cim:releaseDate'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('sub_components', True, decode_model_component, 'child::cim:childComponent/cim:modelComponent'),
        ('sub_components', True, decode_processor_component, 'child::cim:childComponent/cim:processorComponent'),
        ('timing', False, decode_timing, 'child::cim:timing'),
        ('type', False, 'unicode', 'child::cim:type[1]/@value'),
        ('types', True, 'unicode', 'child::cim:type/@value'),
    ]

    return set_attributes(typeset.software.ModelComponent(), xml, nsmap, decodings)


def decode_parallelisation(xml, nsmap):
    """Decodes an instance of the following type: parallelisation.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Parallelisation

    """
    decodings = [
        ('processes', False, 'int', 'child::cim:processes'),
        ('ranks', True, decode_rank, 'child::cim:rank'),
    ]

    return set_attributes(typeset.software.Parallelisation(), xml, nsmap, decodings)


def decode_processor_component(xml, nsmap):
    """Decodes an instance of the following type: processor component.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.ProcessorComponent

    """
    decodings = [
        ('citations', True, decode_citation, 'child::cim:citation'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('language', False, decode_component_language, 'child::cim:componentLanguage'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:modelComponent'),
        ('properties', True, decode_component_property, 'child::cim:componentProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:numericalProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:scientificProperties/cim:componentProperty'),
        ('release_date', False, 'datetime.datetime', 'child::cim:releaseDate'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('sub_components', True, decode_model_component, 'child::cim:childComponent/cim:modelComponent'),
        ('sub_components', True, decode_processor_component, 'child::cim:childComponent/cim:processorComponent'),
    ]

    return set_attributes(typeset.software.ProcessorComponent(), xml, nsmap, decodings)


def decode_rank(xml, nsmap):
    """Decodes an instance of the following type: rank.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Rank

    """
    decodings = [
        ('increment', False, 'int', 'child::cim:rankIncrement'),
        ('max', False, 'int', 'child::cim:rankMax'),
        ('min', False, 'int', 'child::cim:rankMin'),
        ('value', False, 'int', 'child::cim:rankValue'),
    ]

    return set_attributes(typeset.software.Rank(), xml, nsmap, decodings)


def decode_spatial_regridding(xml, nsmap):
    """Decodes an instance of the following type: spatial regridding.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.SpatialRegridding

    """
    decodings = [
        ('dimension', False, 'unicode', 'child::cim:spatialRegriddingDimension'),
        ('properties', True, decode_spatial_regridding_property, 'child::cim:spatialRegriddingProperty'),
        ('standard_method', False, 'unicode', 'child::cim:spatialRegriddingStandardMethod'),
        ('user_method', False, decode_spatial_regridding_user_method, 'child::cim:spatialRegriddingUserMethod'),
    ]

    return set_attributes(typeset.software.SpatialRegridding(), xml, nsmap, decodings)


def decode_spatial_regridding_property(xml, nsmap):
    """Decodes an instance of the following type: spatial regridding property.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.SpatialRegriddingProperty

    """
    decodings = [
        ('name', False, 'unicode', 'child::cim:name'),
        ('value', False, 'unicode', 'child::cim:value'),
    ]

    return set_attributes(typeset.software.SpatialRegriddingProperty(), xml, nsmap, decodings)


def decode_spatial_regridding_user_method(xml, nsmap):
    """Decodes an instance of the following type: spatial regridding user method.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.SpatialRegriddingUserMethod

    """
    decodings = [
        ('file', False, decode_data_object, 'child::cim:file/cim:dataObject'),
        ('file', False, decode_doc_reference, 'child::cim:file/cim:reference'),
        ('name', False, 'unicode', 'child::cim:name'),
    ]

    return set_attributes(typeset.software.SpatialRegriddingUserMethod(), xml, nsmap, decodings)


def decode_statistical_model_component(xml, nsmap):
    """Decodes an instance of the following type: statistical model component.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.StatisticalModelComponent

    """
    decodings = [
        ('citations', True, decode_citation, 'child::cim:citation'),
        ('description', False, 'unicode', 'child::cim:description'),
        ('language', False, decode_component_language, 'child::cim:componentLanguage'),
        ('long_name', False, 'unicode', 'child::cim:longName'),
        ('meta', False, decode_doc_meta_info, 'self::cim:statisticalModelComponent'),
        ('properties', True, decode_component_property, 'child::cim:componentProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:numericalProperties/cim:componentProperty'),
        ('properties', True, decode_component_property, 'child::cim:scientificProperties/cim:componentProperty'),
        ('release_date', False, 'datetime.datetime', 'child::cim:releaseDate'),
        ('responsible_parties', True, decode_responsible_party, 'child::cim:responsibleParty'),
        ('short_name', False, 'unicode', 'child::cim:shortName'),
        ('sub_components', True, decode_model_component, 'child::cim:childComponent/cim:modelComponent'),
        ('sub_components', True, decode_processor_component, 'child::cim:childComponent/cim:processorComponent'),
        ('timing', False, decode_timing, 'child::cim:timing'),
        ('type', False, 'unicode', 'child::cim:type[1]/@value'),
        ('types', True, 'unicode', 'child::cim:type/@value'),
    ]

    return set_attributes(typeset.software.StatisticalModelComponent(), xml, nsmap, decodings)


def decode_time_lag(xml, nsmap):
    """Decodes an instance of the following type: time lag.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.TimeLag

    """
    decodings = [
        ('units', False, 'unicode', '@units'),
        ('value', False, 'int', 'child::cim:value'),
    ]

    return set_attributes(typeset.software.TimeLag(), xml, nsmap, decodings)


def decode_time_transformation(xml, nsmap):
    """Decodes an instance of the following type: time transformation.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.TimeTransformation

    """
    decodings = [
        ('description', False, 'unicode', 'child::cim:description'),
        ('mapping_type', False, 'unicode', 'child::cim:mappingType/@value'),
    ]

    return set_attributes(typeset.software.TimeTransformation(), xml, nsmap, decodings)


def decode_timing(xml, nsmap):
    """Decodes an instance of the following type: timing.

    :param lxml.etree xml: XML from which type is to be decoded.
    :param dict nsmap: XML namespace mappings.

    :returns: A decoded type instance.
    :rtype: cim.v1.typeset.software.Timing

    """
    decodings = [
        ('end', False, 'datetime.datetime', 'child::cim:end'),
        ('is_variable_rate', False, 'bool', '@variableRate'),
        ('rate', False, 'int', 'child::cim:rate'),
        ('start', False, 'datetime.datetime', 'child::cim:start'),
        ('units', False, 'unicode', '@units'),
    ]

    return set_attributes(typeset.software.Timing(), xml, nsmap, decodings)





# -*- coding: utf-8 -*-

"""
.. module:: setup.py

   :license: GPL / CeCILL
   :platform: Unix
   :synopsis: pyesdoc setup.

.. moduleauthor:: ES-DOC <dev@esdocumentation.org>

"""
import os
import re
from codecs import open

from setuptools import setup
from setuptools import find_packages
from setuptools.dist import Distribution


# List of 3rd party python dependencies.
_REQUIRES = [
    'arrow',
    'feedparser',
    'lxml',
    'nose',
    'requests',
    'tornado',
    'jupyter'
    ]


class _BinaryDistribution(Distribution):
    """Distribution sub-class to override defaults.

    """
    def is_pure(self):
        """Gets flag indicating whether build is pure python or not.

        """
        return False


def _read(fname):
    """Returns content of a file.

    """
    fpath = os.path.dirname(__file__)
    fpath = os.path.join(fpath, fname)
    with open(fpath, 'r', 'utf-8') as file_:
        return file_.read()


def _get_version():
    """Returns library version by inspecting __init__.py file.

    """
    return re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                     _read("pyesdoc/__init__.py"),
                     re.MULTILINE).group(1)


# Libary version.
_VERSION = _get_version()

# Library packages.
_PACKAGES = find_packages()

# User readme.
_README = _read('README.rst')



setup(
    name='pyesdoc',
    version=_VERSION,
    description='pyesdoc is a python library for publishing earth system documentation.',
    long_description=_README,
    author='Mark A. Greenslade',
    author_email='momipsl@ipsl.jussieu.fr',
    url='https://github.com/ES-DOC/esdoc-py-client',
    packages=_PACKAGES,
    include_package_data=True,
    install_requires=_REQUIRES,
    license='GPL/CeCILL-2.1',
    zip_safe=False,
    distclass=_BinaryDistribution,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface carbon cycle'

# --------------------------------------------------------------------
# CARBON CYCLE: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Properties of land surface carbon cycle',
    'properties': [
        ('tiling', 'str', '0.1',
             'Describe the carbon cycle tiling, if any.'),
        ('time_step', 'int', '1.1',
             'Time step of carbon cycle in seconds'),
        ('anthropogenic_carbon', 'ENUM:anthropogenic_carbon_methods', '0.N',
             'Describe the treament of the anthropogenic carbon pool'),
        ('prognostic_variables', 'str', '1.1',
             'List the prognostic variables of the carbon scheme'),

    ]
}

# --------------------------------------------------------------------
# CARBON CYCLE: process
# --------------------------------------------------------------------
DETAILS['vegetation'] = {
    'description': 'TODO',
    'properties' : [ 
        ('number_of_carbon_pools', 'int', '1.1',
             'Enter the number of carbon pools used'),
        ('carbon_pools', 'str', '0.1',
             'List the carbon pools used'),
        ('forest_stand_dynamics', 'str', '0.1',
             'Describe the treatment of forest stand dyanmics'),
    ],
    'detail_sets': [
        'photosynthesis',
        'autotrophic_respiration',
        'allocation',
        'phenology',
        'mortality',
    ]
}

DETAILS['vegetation:photosynthesis'] = {
    'description': 'TODO',
    'properties' : [
        ('method', 'str', '0.1',
            'Describe the general method used for photosynthesis (e.g. type of photosynthesis, distinction between C3 and C4 grasses, Nitrogen depencence, etc.)',)
    ]
}

DETAILS['vegetation:autotrophic_respiration'] = {
    'description': 'TODO',
    'properties' : [
        ('maintainance_respiration', 'str', '0.1',
            'Describe the general method used for maintainence respiration'),
        ('growth_respiration', 'str', '0.1',
            'Describe the general method used for growth respiration'),
        ]
    }

DETAILS['vegetation:allocation'] = {
    'description': 'TODO',
    'properties' : [
        ('method', 'str', '1.1',
             'Describe the general principle behind the allocation scheme'),
        ('allocation_bins', 'ENUM:allocation_bin_types', '1.1',
            'Specify distinct carbon bins used in allocation'),
        ('allocation_fractions', 'ENUM:allocation_fraction_types', '1.1',
            'Describe how the fractions of allocation are calculated'),
        ]
    }

DETAILS['vegetation:phenology'] = {
    'description': 'TODO',
    'properties' : [
        ('method', 'str', '1.1',
             'Describe the general principle behind the phenology scheme'),
    ]
}

DETAILS['vegetation:mortality'] = {
    'description': 'TODO',
    'properties' : [
        ('method', 'str', '1.1',
             'Describe the general principle behind the mortality scheme'),
    ]
}

# --------------------------------------------------------------------
# CARBON CYCLE: process
# --------------------------------------------------------------------
DETAILS['litter'] = {
    'description': 'TODO',
    'properties' : [
        ('number_of_carbon_pools', 'int', '1.1',
             'Enter the number of carbon pools used'),
        ('carbon_pools', 'str', '0.1',
             'List the carbon pools used'),
        ('decomposition', 'str', '0.1',
             'List the decomposition methods used'),
        ('method', 'str', '0.1',
             'List the general method used'),
    ]
}

# --------------------------------------------------------------------
# CARBON CYCLE: process
# --------------------------------------------------------------------
DETAILS['soil'] = {
    'description': 'TODO',
    'properties' : [
        ('number_of_carbon_pools', 'int', '1.1',
             'Enter the number of carbon pools used'),
        ('carbon_pools', 'str', '0.1',
             'List the carbon pools used'),
        ('decomposition', 'str', '0.1',
             'List the decomposition methods used'),
        ('method', 'str', '0.1',
             'List the general method used'),
    ]
}


# --------------------------------------------------------------------
# CARBON CYCLE: process
# --------------------------------------------------------------------
DETAILS['permafrost_carbon'] = {
    'description': 'TODO',
    'properties' : [
        ('is_permafrost_included', 'bool', '1.1',
             'Is permafrost included?'),
        ('emitted_greenhouse_gases', 'str', '0.1',
             'List the GHGs emitted'),
        ('decomposition', 'str', '0.1',
             'List the decomposition methods used'),
        ('impact_on_soil_properties', 'str', '0.1',
             'Describe the impact of permafrost on soil properties'),
    ]
}

# --------------------------------------------------------------------
# CARBON CYCLE: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['anthropogenic_carbon_methods'] = {
    'description': 'Treatment of the anthropogenic carbn pool',
    'is_open': True,
    'members': [
        ('grand slam protocol', None),
        ('residence time', None),
        ('decay time', None),
    ]
}

ENUMERATIONS['allocation_bin_types'] = {
    'description': 'Specify the allocation of vegetation carbon bins',
    'is_open': True,
    'members': [
        ('leaves + stems + roots', None),
        ('leaves + stems + roots (leafy + woody)', None),
        ('leaves + fine roots + coarse roots + stems', None),
        ('whole plant (no distinction)', None),
        ]
    }

ENUMERATIONS['allocation_fraction_types'] = {
    'description': 'Describe how the fractions of allocation are calculated',
    'is_open': True,
    'members': [
        ('fixed', None),
        ('function of vegetation type', None),
        ('function of plant allometry', None),
        ('explicitly calculated', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: data_classes.py
   :synopsis: Set of CIM v2 ontology type definitions.

"""


def data_association_types():
    """Set of possible dataset associations.

    Selected from, and extended from,  ISO19115 (2014) DS_AssociationTypeCode.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ("revisonOf", "This dataset was revised from the target"),
            ("partOf", "This dataset forms part of the target"),
            ("isComposedOf", "This dataset is composed from the target")
        ]
    }


def dataset():
    """Dataset discovery information.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'is_document': True,
        'pstr': ('{}', ('name',)),
        'properties': [
            ('citations', 'shared.citation', '0.N',
                "Set of pertinent citations."),
            ('responsible_parties', 'linked_to(shared.responsibility)', '0.N',
                "Individuals and organisations reponsible for the data."),

            ('name', 'str', '1.1',
                "Name of dataset."),
            ('description', 'str', '0.1',
                "Textural description of dataset."),

            ('availability', 'shared.online_resource', '0.N',
                "Where the data is located, and how it is accessed."),
            ('drs_datasets', 'drs.drs_publication_dataset', '0.N',
                "Data available in the DRS."),
            ('produced_by', 'linked_to(data.simulation)', '0.1',
                "Makes a link back to originating activity."),
            ('related_to_dataset', 'shared.online_resource', '0.N',
                "Related dataset."),
        ]
    }

def input_dataset():
    """An input dataset is used as within another component (such as a
model). It comprises an original, source dataset plus any
modifications requirted to use it in the relevant component.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('original_data', 'data.dataset', '1.1',
                'The source dataset, prior to any modifications'),
            ('modifications_applied', 'str', '0.1',
                'Describe modifications (if any) applied to the dataset prior to use. E.g. spatial interpolation, temporal averaging, etc.'),
        ]
    }


def downscaling():
    """Defines a downscaling activity.

    """
    return {
        'type': 'class',
        'base': 'data.simulation',
        'is_abstract': False,
        'properties': [
            ('downscaled_from', 'linked_to(data.simulation)', '1.1',
                "The simulation that was downscaled by this downscaling activity.")
        ],
        'constraints': [
            ('cardinality', 'parent_simulation', '0.0')
        ]
    }


def simulation():
    """Simulation class provides the integrating link about what models
    were run and wny.  In many cases this should be auto-generated
    from output file headers.

    """

    return {
        'type': 'class',
        'base': 'activity.activity',
        'is_abstract': False,
        'properties': [
            ('part_of_project', 'linked_to(designing.project)', '1.N',
                'Project or projects for which simulation was run'),
            ('ran_for_experiments', 'linked_to(designing.numerical_experiment)', '1.N',
                'One or more experiments with which the simulation is associated'),
            ('sub_experiment', 'designing.numerical_experiment', '0.1',
                "For start-date ensembles, this will indicate the beginning year; for offline models driven by output from another model, this will provide the source_id and variant_label for the 'driving' model."),
            ('used', 'linked_to(science.model)', '1.1',
                'The model used to run the simulation'),
            ('primary_ensemble', 'linked_to(activity.ensemble)', '0.1',
                'Primary Ensemble (ensemble for which this simulation was first run).'),
            ('insitution', 'shared.party', '0.1',
                'institution which carried out the simulation'),
            ('parent_simulation', 'activity.parent_simulation', '0.1',
                'If appropriate, detailed information about how this simulation branched from a previous one'),

            # Time
            ('start_time', 'time.date_time', '0.1',
                'The start date-time of the simulation. e.g. 2012-04-01 00:00:00'),
            ('end_time', 'time.date_time', '0.1',
                'The start date-time of the simulation. e.g. 2087-11-30 12:00:00'),
            ('calendar', 'time.calendar', '0.1',
                'The calendar used in the simulation'),

            # Ensemble member attributes
            ('realization_index', 'int', '0.1',
                'realization number, e.g. 5'),
            ('initialization_index', 'int', '0.1',
                'Index variant of initialization method, e.g. 1'),
            ('physics_index', 'int', '0.1',
                'index for model physics, e.g. 3'),
            ('forcing_index', 'int', '0.1',
                'index for variant of forcing, e.g. 2'),
            ('variant_info', 'str', '0.1',
                'description of run variant differences, e.g. forcing: black carbon aerosol only'),

            # Further Info URL
            ('further_info_url', 'str', '0.1',
                'On-line location of documentation'),

            # Extra attributes
            ('extra_attributes','shared.extra_attribute', '0.N',
                'Additional attributes provided with simulation.')
            ],

        'constraints':[
            ('cardinality', 'rationale', '0.0'),
        ]
    }

def variable_collection():
    """A collection of variables within the scope of a code or process element.

    """
    return {
        'type': 'class',
        'base': None,
        'is_abstract': False,
        'properties': [
            ('collection_name', 'str', '0.1',
                "Name for this variable collection."),
            ('variables', 'str', '1.N',
                "Set of variable names.")
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: activity_numerical_experiment.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.activity.numerical_experiment document.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def get_extenders():
    """Returns set of extension functions.

    """
    return (
        _set_full_experiment_name,
        )


def _set_full_experiment_name(ctx):
    """Sets full experiment name.

    """
    name = ""
    if ctx.doc.experiment_id is not None:
    	name += ctx.doc.experiment_id
    	name += " "
    name += ctx.doc.short_name

    ctx.ext.full_experiment_name = name




"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface soil'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'TODO',
    'properties': [
         ('heat_water_coupling', 'str', '1.1',
             'Describe the coupling between heat and water in the soil'),
         ('number_of_soil layers', 'int', '1.1',
             'The number of soil layers'),
         ('prognostic_variables', 'str', '1.1',
             'List the prognostic variables of the soil scheme'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: Soil map
# --------------------------------------------------------------------
DETAILS['soil_map'] = {
    'description': 'Key properties of the land surface soil map',
    'properties': [
        ('description', 'str', '1.1',
             'General description of soil map'),
        ('structure', 'str', '0.1',
             'Describe the soil structure map'),
        ('texture', 'str', '0.1',
             'Describe the soil texture map'),
        ('organic_matter', 'str', '0.1',
             'Describe the soil organic matter map'),
        ('albedo', 'str', '0.1',
             'Describe the soil albedo map'),
        ('water_table', 'str', '0.1',
             'Describe the soil water table map, if any'),
        ('continuously_varying_soil_depth', 'bool', '1.1',
             'Does the soil properties vary continuously with depth?'),
        ('soil_depth', 'str', '0.1',
             'Describe the soil depth map'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Snow-free albedo
# --------------------------------------------------------------------
DETAILS['snow_free_albedo'] = {
    'description': 'TODO',
    'properties' : [
        ('prognostic', 'bool', '1.1',
             'Is snow free albedo prognostic?'),
        ('functions', 'ENUM:snow_free_albedo_function_types', '0.N',
             'If prognostic, describe the dependancies on snow free albedo calculations'),
        ('direct_diffuse', 'ENUM:direct_diffuse_types', '0.1',
             'If prognostic, describe the distinction between direct and diffuse albedo'),
        ('number_of_wavelength_bands', 'int', '0.1',
             'If prognostic, enter the number of wavelength bands used'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Hydrology
# --------------------------------------------------------------------
DETAILS['hydrology'] = {
    'description': 'Key properties of the land surface soil hydrology',
    'properties': [
        ('description', 'str', '1.1',
             'General description of the soil hydrological model'),
        ('time_step', 'int', '1.1',
             'Time step of river soil hydrology in seconds'),                
        ('tiling', 'str', '0.1',
             'Describe the soil hydrology tiling, if any.'),
        ('vertical_discretisation', 'str' , '1.1',
             'Describe the typical vertical discretisation'),
        ('number_of_ground_water_layers', 'int', '1.1',
             'The number of soil layers that may contain water'),
        ('lateral_connectivity', 'ENUM:lateral_connectivity_types', '1.N',
             'Describe the lateral connectivity between tiles'),
        ('method', 'ENUM:water_storage_method_types', '1.1',
             'The hydrological dynamics scheme in the land surface model'),
    ],
    'detail_sets': [
        'freezing',
        'drainage'
    ]
}

DETAILS['hydrology:freezing'] = {
    'description': 'TODO',
    'properties': [
        ('number_of_ground_ice_layers', 'int', '1.1',
            'How many soil layers may contain ground ice'),
        ('ice_storage_method', 'str', '1.1',
            'Describe the method of ice storage'),
        ('permafrost', 'str', '1.1',
            'Describe the treatment of permafrost, if any, within the land surface scheme'),
        ]
    }

DETAILS['hydrology:drainage'] = {
    'description': 'TODO',
    'properties': [
        ('description', 'str', '1.1',
            'General describe how drainage is included in the land surface scheme'),
        ('types', 'ENUM:drainage_types', '0.N',
            'Different types of runoff represented by the land surface model'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: Heat treatment
# --------------------------------------------------------------------
DETAILS['heat_treatment'] = {
    'description': 'TODO',
    'properties': [
        ('description', 'str', '1.1',
             'General description of how heat treatment properties are defined'),
        ('time_step', 'int', '1.1',
             'Time step of soil heat scheme in seconds'),                
        ('tiling', 'str', '0.1',
             'Describe the soil heat treatment tiling, if any.'),
        ('vertical_discretisation', 'str' , '1.1',
             'Describe the typical vertical discretisation'),
        ('heat_storage', 'ENUM:heat_storage_types', '1.1',
             'Specify the method of heat storage'),
        ('processes', 'ENUM:heat_treatment_process_types', '1.N',
             'Describe processes included in the treatment of soil heat'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['lateral_connectivity_types'] = {
    'description': 'Describe the lateral connectivity between tiles',
    'is_open': True,
    'members': [
        ('perfect connectivity', 'Common soil for multiple tiles'),
        ('Darcian flow', 'Darcian flow among hillslope tiles'),
    ]
}

ENUMERATIONS['drainage_types'] = {
    'description': 'Different types of runoff represented by the land surface model',
    'is_open': True,
    'members': [
        ('Gravity drainage', None),
        ('Horton mechanism', None),
        ('topmodel-based', None),
        ('Dunne mechanism', None),
        ('Lateral subsurface flow', None),
        ('Baseflow from groundwater', None),
    ]
}

ENUMERATIONS['water_storage_method_types'] = {
    'description': 'Describe the method by which water is stored in the land surface scheme/model',
    'is_open': True,
    'members': [
        ('Bucket', None),
        ('Force-restore', None),
        ('Choisnel', None),
        ('Explicit diffusion', None),
        ]
    }

ENUMERATIONS['heat_storage_types'] = {
    'description': 'Method of heat storage',
    'is_open': True,
    'members': [
        ('Force-restore', None),
        ('Explicit diffusion', None),
        ]
    }

ENUMERATIONS['heat_treatment_process_types'] = {
    'description': 'Processes included in the treatment of soil heat',
    'is_open': True,
    'members': [
        ('soil moisture freeze-thaw', None),
        ('coupling with snow temperature', None),
    ]
}
    
ENUMERATIONS['snow_free_albedo_function_types'] = {
    'description': 'Describe the dependancies on snow free albedo calculations',
    'is_open': True,
    'members': [
        ('vegetation type', None),
        ('soil humidity', None),
        ('vegetation state', None),        
    ]
}
    
ENUMERATIONS['direct_diffuse_types'] = {
    'description': 'Describe the distinction between direct and diffuse albedo',
    'is_open': True,
    'members': [
        ('distinction between direct and diffuse albedo', None),
        ('no distinction between direct and diffuse albedo', None),        
    ]
}
    
